
/*==============================================================================
 *
 * (#)Xmame.java
 * created   2002/11/27
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *==============================================================================
 */

package com.miami.romeo.xmame;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.ArrayList;
import java.net.*;

import javax.swing.*;

import com.miami.romeo.main.Romeo;
import com.miami.romeo.config.*;

/**
 * Xmame.<br>
 * This class dialogs with the xmame executable. It can retrieve
 * data (ie list of games) and run a game.
 *
 * @author    $author: Stefano Bolli
 * @version   $Revision: 1.8 $Date: 2004/01/04 00:43:09 $
 */
public class Xmame {
  //******* Classes needed by Xmame.java
  Romeo romeo = null;
  Utils utils = new Utils();
  Config config = null;
  private String romsInfoFilePath;
  private File romsInfoFile;
  private File config_file;

  // Yes and No :|
  String yes = "Yes";
  String no  = "No";
  String cmd;

  // Array containing output list generated by xmame commands
  String outlist[];

  // Paths to xmame samples dir
  private String samplesDirPath;

  /**
   * Constructor
   *
   * @param romeo The main class
   */
  public Xmame(Romeo romeo) {
    this.romeo          = romeo;

    // Configuration file
    String file_name    = System.getProperty("user.home") + File.separator +
                          ".romeo4xmame" + File.separator + "Romeo.ini";
    this.config_file    = new File(file_name);
    
    // Class to read the configuration file
    config              = new Config(config_file);
    
    // Paths to any dirs and to catver.ini
    this.samplesDirPath = config.get("xmame.sample_dir");
  }

  /**
   * This method retrieves list of supported games with full description
   * running the command "xmame -listfull".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of supported games with full description
   *
   * @see             #runCommand(String)
   */
  public String[] listFull(String xmamePath) {
    ArrayList listFull = new ArrayList();

    try {
      // Run command 'xmame -listfull'
      cmd = getRealCommand(xmamePath + " -listfull -sortname");
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
	listFull.add(outlist[i]);
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listFull.toArray(new String[0]);
  }

  /**
   * This method retrieves game names.
   *
   * @param listFull The list of supported games with full description
   *
   * @return         The list of game names
   *
   * @see            #listFull(String)
   */
  public String[] getRomNames(String[] listFull) {
    String romName;
    ArrayList romNameList = new ArrayList();

    for (int i = 0; i < listFull.length; i++) {
      int qm_idx = listFull[i].indexOf('"');

      if (qm_idx > 0) {
        romName = listFull[i].substring(0, qm_idx).trim();

	romNameList.add(romName);
      }
    }

    return (String[])romNameList.toArray(new String[0]);
  }

  /**
   * This method returns an array of favorite/non-favorite games.
   *
   * @param romName   The game names
   * @param listPath The path to the list of games file
   *
   * @return         The list of favorite/non-favorite games
   *
   * @see            #getRomNames(String[])
   */
  public String[] getFavoriteRoms(String[] romName, String listPath) {
    String favoriteRom;
    utils = new Utils();
    ArrayList favoriteRomList = new ArrayList();

    String config_name = System.getProperty("user.home") + File.separator +
                         ".romeo4xmame" + File.separator + "favorites" + 
			 listPath.substring(4) + ".ini";
    File config_file = new File(config_name);
    config = new Config(config_file);

    for (int i = 0; i < romName.length; i++) {
      String favorite = config.get(romName[i]);

      if ((favorite != null) && (favorite.equals(yes)))
        favoriteRom = yes;
      else
        favoriteRom = no;

      favoriteRomList.add(favoriteRom);
    }

    return (String[])favoriteRomList.toArray(new String[0]);
  }

  /**
   * This method retrieves game descriptions.
   *
   * @param listFull The list of supported games with full description
   *
   * @return         The list of descriptions
   *
   * @see            #listFull(String)
   */
  public String[] getRomDescriptions(String[] listFull) {
    String line, romDescription;
    ArrayList romDescriptionList = new ArrayList();
    int qm_idx;

    for (int i = 0; i < listFull.length; i++) {
      qm_idx = listFull[i].indexOf('"');

      if (qm_idx > 0) {
	// Parse line
	line = listFull[i].substring(qm_idx + 1).trim();
	qm_idx = line.indexOf('"');

	if (qm_idx > 0) {
         romDescription = line.substring(0, qm_idx).trim();

         romDescriptionList.add(romDescription);
        }
      }
    }

    return (String[])romDescriptionList.toArray(new String[0]);
  }

  /**
   * This method retrieves the list of supported games with
   * manufacturer and year running the command "xmame -listgames".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of supported games with
   *                  manufacturer and year
   *
   * @see             #runCommand(String)
   */
  public String[] listGames(String xmamePath) {
    ArrayList listOfGames = new ArrayList();

    try {
      cmd = getRealCommand(xmamePath + " -listgames -sortname");
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
	listOfGames.add(outlist[i]);
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listOfGames.toArray(new String[0]);
  }

  /**
   * This method retrieves the list of the production year
   * for each game.
   *
   * @param listOfGames The list of supported games with
   *                    manufacturer and year.
   *
   * @return            The list of the production years
   *
   * @see               #listGames(String)
   */
  public String[] getRomYears(String[] listOfGames) {
    String romYear;
    ArrayList yearList = new ArrayList();

    for (int i = 0; i < listOfGames.length; i++) {
      if ((listOfGames[i].startsWith("19")) || 
          (listOfGames[i].startsWith("20")) ||
          (listOfGames[i].startsWith("????")))  {
        romYear = listOfGames[i].substring(0, 4).trim();

	yearList.add(romYear);
      }
    }

    return (String[])yearList.toArray(new String[0]);
  }

  /**
   * This method retrieves the manufacturer for each game.
   *
   * @param listOfGames The list of supported games with
   *                    manufacturer and year.
   *
   * @return            The list of manufacturers
   *
   * @see               #listGames(String)
   */
  public String[] getRomManufacturers(String[] listOfGames) {
    String romManufacturer;
    ArrayList manufacturerList = new ArrayList();

    for (int i = 0; i < listOfGames.length; i++) {
      if (listOfGames[i].startsWith("19") ||
          listOfGames[i].startsWith("20") ||
          listOfGames[i].startsWith("????"))  {
        romManufacturer = listOfGames[i].substring(4, 41).trim();

	manufacturerList.add(romManufacturer);
      }
    }

    return (String[])manufacturerList.toArray(new String[0]);
  }

  /**
   * This method runs the command "xmame -listgamelist" that
   * outputs a list with specialy formatted extra info for
   * generating gamelist.mame. Next this void retrieves if
   * each game works fine or doesn' t. Finally it puts all
   * retrieved data into an array.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          Array of strings
   *
   * @see             #runCommand(String)
   */
  public String[] getWorkingRoms(String xmamePath) {
    String line, working;
    ArrayList workingRomList = new ArrayList();

    try {
      String OS = System.getProperty("os.name").toLowerCase();
      // System.out.println(OS);
      if (OS.indexOf("windows") > -1) {
        cmd = getRealCommand(xmamePath + " -gamelist -sortname");
      }
      else {
        cmd = xmamePath + " -listgamelist -sortname";
      }

      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
        // Remove all unuseful things...
	if ((!outlist[i].startsWith("| Game Name")) &&
	    (!outlist[i].endsWith("Internal |")) &&
	    (outlist[i].startsWith("|")))  {
          line = outlist[i].substring(36, 43).trim();

	  if (line.startsWith(yes))
            working = yes;
          else
            working = no;

	  workingRomList.add(working);
	}
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])workingRomList.toArray(new String[0]);
  }

  /**
   * This method retrieves the list of all clones running
   * the command "xmame -listclones".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of all clones
   *
   * @see             #runCommand(String)
   */
  public String[] listClones(String xmamePath) {
    String romCloneOf;
    ArrayList romsCloneOfList = new ArrayList();

    try {
      // Run command "xmame -listclones"
      cmd = getRealCommand(xmamePath + " -listclones -sortname");
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
        // Do not add other things as copyright lines (Open GL) or....
        if (!outlist[i].startsWith("Name") &&
	    // Needed by xmame.xgl --> thanks to Antonio Giungato
	    !outlist[i].startsWith("GLmame") &&
	    !outlist[i].startsWith("thel") &&
            !outlist[i].startsWith("based") &&
            !outlist[i].equals("")) {
          romCloneOf = outlist[i].trim();

          romsCloneOfList.add(romCloneOf);
	}
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])romsCloneOfList.toArray(new String[0]);
  }
  
  /**
   * This method retrieves the clones of a given game.
   *
   * @param xmamePath The path to the xmame executable
   * @param gameName  The game name
   *
   * @return          The list of the clones of a given game
   *
   * @see             #runCommand(String)
   */
  public String[] getClonesOfGame(String xmamePath, String gameName) {
    ArrayList cloneOfList = new ArrayList();

    try {
      // Run command "xmame -listclones game_name"
      cmd = getRealCommand(xmamePath + " -listclones -sortname " + gameName);
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
        // Do not add other things as copyright lines (Open GL) or....
        if (// Needed by xmame.xgl --> thanks to Antonio Giungato
	    !outlist[i].startsWith("GLmame") && 
	    !outlist[i].startsWith("thel") &&
            !outlist[i].startsWith("based") &&
            !outlist[i].equals("")) {
	    
	  if (outlist[i].startsWith("Name:"))
            cloneOfList.add("Name:\tClone of:");
          else
            cloneOfList.add(outlist[i].substring(0, 9).trim() + "\t" +
	                    outlist[i].substring(9).trim());
	}
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])cloneOfList.toArray(new String[0]);
  }

  /**
   * This method compares the list of games with the list of clones.
   * 
   * @param romName       The game names
   * @param listOfClones  The clones
   *
   * @return              The list of original and cloned games
   *
   * @see                 #getRomNames(String[])
   */
  public String[] getClones(String[] romName, String[] listOfClones) {
    String cloneof;
    String[] clone  = new String[romName.length];

    for (int i = 0; i < romName.length; i++) {
      clone[i] = "-";

      for (int j = 0; j < listOfClones.length; j++) {
        if (romName[i].equals(listOfClones[j].substring(0, 9).trim())) {
          clone[i] = listOfClones[j].substring(9).trim();
	  break;
	}
      }
    }

    return clone;
  }

  /**
   * This method retrieves list of source files running the command
   * "xmame -lsf".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of source files
   *
   * @see             #runCommand(String)
   */
  public String[] getSourceFiles(String xmamePath) {
    String sourceFile;
    ArrayList listOfSourceFiles = new ArrayList();

    try {
      cmd = getRealCommand(xmamePath + " -listsourcefile -sortname"); 
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
        if (outlist[i].endsWith(".c")) {
          int idx_separator = outlist[i].lastIndexOf('/');
	  int idx_extension = outlist[i].indexOf(".c");
	  sourceFile = outlist[i].substring(idx_separator + 1, idx_extension);
	  listOfSourceFiles.add(sourceFile);
	}
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listOfSourceFiles.toArray(new String[0]);
  }

  /**
   * This method retrieves the list of games that use audio samples.
   *
   * @param xmamePath The path to the xmame executable
   * @param romName   The list of rom names
   *
   * @return          The list of games that use audio samples
   *
   * @see             #listSamples(String)
   */
  public String[] getRomsUsingSamples(String xmamePath, String[] romName) {
    ArrayList listOfRomsUsingSamples = new ArrayList();
    String sample = "No";

    String OS = System.getProperty("os.name").toLowerCase();
    // System.out.println(OS);
    if (OS.indexOf("windows") > -1) { 
      // Windows port doesn' t support this option...
      for (int i = 0; i < romName.length; i++) {
        listOfRomsUsingSamples.add("-");
      }
    }
    else {
      String[] samples = listSamples(xmamePath);

      for (int i = 0; i < romName.length; i++) {
        for (int j = 0; j < samples.length; j++) {
	  if (samples[j].equals("*" + romName[i])) {
            if (!samplesDirPath.equals("")) {
	      File sampleFile = new File(samplesDirPath + romName[i] + ".zip");

	      if (sampleFile.exists()) {
	        sample = "Yes: available";
	      }
	      else {
	        sample = "Yes: not available";
	      }
	    }
	    else {
	      sample = "Yes";
	    }
          }
	
	  break;
        } 

        listOfRomsUsingSamples.add(sample);
      }
    }

    return (String[])listOfRomsUsingSamples.toArray(new String[0]);
  }
  
  /**
   * This method lists used audio samples running the command 
   * "xmame -ls".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of used audio samples
   */
  public String[] listSamples(String xmamePath) {
    cmd = new String();
    ArrayList listSamples = new ArrayList();

    try {
      // Run command 'xmame -ls'
      cmd = getRealCommand(xmamePath) + " -ls";
      
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
	listSamples.add(outlist[i].trim());
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listSamples.toArray(new String[0]);
  }
  
  /**
   * This method lists used audio samples for a given game running
   * the command "xmame -ls gamename".
   *
   * @param xmamePath The path to the xmame executable
   * @param gameName  The game name
   *
   * @return          The list of used audio samples for a given name
   *
   * @see             #runCommand(String)
   */
  public String[] getGameSamples(String xmamePath, String gameName) {
    cmd = new String();
    ArrayList listSamples = new ArrayList();

    try {
      String OS = System.getProperty("os.name").toLowerCase();
      // System.out.println(OS);
      if (OS.indexOf("windows") > -1)
        cmd = getRealCommand(xmamePath) + " " + gameName + " -listsamples";
      else
        // Run command 'xmame -ls'
        cmd = xmamePath + " -ls " + gameName;

      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
	if (outlist[i].endsWith(".wav"))
	  listSamples.add(outlist[i].trim());
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listSamples.toArray(new String[0]);
  }
  
  /**
   * This method retrieves detailed info for a given
   * game running the command "xmame -ld gamename".
   *
   * @param xmamePath The path to the xmame executable
   * @param gameName  The game name
   *
   * @return          The detailed info for a given game
   *
   * @see             #runCommand(String)
   */
  public String getDetailedInfo(String xmamePath, String gameName) {
    String detailedInfo = new String();

    try {
      String OS = System.getProperty("os.name").toLowerCase();
      // System.out.println(OS);
      if (OS.indexOf("windows") > -1)
        cmd = getRealCommand(xmamePath + " -listdetails " + gameName);
      else
        // Run command 'xmame -ld'
        cmd = xmamePath + " -ld " + gameName;
      
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
	if (outlist[i].startsWith(gameName)) {
	  detailedInfo = outlist[i]; /*.substring(19, 152)*/
	  //System.out.println(outlist[i].substring(19, 152));
	  break;
	}
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return detailedInfo;
  }

  /**
   * This method returns an array with numbers of played
   * matches for each game.
   *
   * @param romName  The list of games
   * @param listPath The path to the list of games
   *
   * @return         An array with the numbers of played matches for each game.
   */
  public String[] getNumberOfPlayedMatches(String[] romName, String listPath) {
    String number;
    utils = new Utils();
    ArrayList listOfNumbers = new ArrayList();

    String config_name = System.getProperty("user.home") + File.separator +
                         ".romeo4xmame" + File.separator + "played" + 
			 listPath.substring(4) + ".ini";
    File config_file = new File(config_name);
    config = new Config(config_file);

    for (int i = 0; i < romName.length; i++) {
      String n = config.get(romName[i]);

      if (n != null)
        number = n;
      else
        number = "0";

      listOfNumbers.add(number);
    }

    return (String[])listOfNumbers.toArray(new String[0]);
  }

  /**
   * This method lists cpu usage statistics per year running
   * the command "xmame -listcpu".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of cpu usage statistics per year
   *
   * @see             #runCommand(String)
   */
  public String[] listCpuUsage(String xmamePath) {
    String cusy;
    ArrayList listCpuUsage = new ArrayList();

    try {
      // Run command "xmame -listcpu"
      String OS = System.getProperty("os.name").toLowerCase();
      // System.out.println(OS);
      if (OS.indexOf("windows") > -1) {
        cmd = getRealCommand(xmamePath + " -listcpuclass");
      }
      else {
        cmd = xmamePath + " -listcpu";
      }

      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++) {
        cusy = outlist[i].trim();

        listCpuUsage.add(cusy);
      }
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])listCpuUsage.toArray(new String[0]);
  }

  /**
   * This method returns xmame version numbers.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @see             #getVersion(String)
   *
   * @return          The major revision number,
   *                  the minor revision number and
   *                  the patch level
   *
   * @since           0.1.3pre2
   */
  public int[] getVersionNumbers(String xmamePath) {
    int[] versionNumber = new int[3];

    String OS = System.getProperty("os.name").toLowerCase();
    // System.out.println(OS);
    if (OS.indexOf("windows") > -1) {
      // Retrieve Windows command line version
      String version = getVersion(xmamePath);
      
      // Identify release
      String release = version.substring(10,
                                         version.lastIndexOf("(")).trim();
      // Put the major revision number into the array
      versionNumber[0] = Integer.parseInt(release.substring(0,
                                          release.indexOf('.')));

      int first_dot = release.indexOf('.');
      int last_dot  = release.lastIndexOf('.');

      if (last_dot > first_dot) {
        // Put the minor revision number into the array
        versionNumber[1] = Integer.parseInt(release.substring(release.indexOf('.') +  1,
                                          release.lastIndexOf('.')));
        // Put the patch level into the array
        versionNumber[2] = Integer.parseInt(release.substring(release.lastIndexOf('.') +  1));
      }
      else {
        versionNumber[1] = Integer.parseInt(release.substring(release.indexOf('.') +  1));
        versionNumber[2] = 0;
      }
    }
    else {
      // Retrieve xmame version
      String version = getVersion(xmamePath);
      // Identify release
      String release = version.substring(version.indexOf("version") + 7,
                                       version.lastIndexOf("(")).trim();
      // Put the major revision number into the array
      versionNumber[0] = Integer.parseInt(release.substring(0,
                                        release.indexOf('.')));

      int first_dot = release.indexOf('.');
      int last_dot  = release.lastIndexOf('.');

      if (last_dot > first_dot) {
        // Put the minor revision number into the array
        versionNumber[1] = Integer.parseInt(release.substring(release.indexOf('.') +  1,
                                          release.lastIndexOf('.')));
        // Put the patch level into the array
        versionNumber[2] = Integer.parseInt(release.substring(release.lastIndexOf('.') +  1));
      }
      else {
        versionNumber[1] = Integer.parseInt(release.substring(release.indexOf('.') +  1));
        versionNumber[2] = 0;
      }
    }

    return versionNumber;
  }

  /**
   * This method retrieves xmame version running the command
   * "xmame -V".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          xmame version
   *
   * @see             #runCommand(String)
   */
  public String getVersion(String xmamePath) {
    File xmame = new File(xmamePath);
    String version = "Romeo 0.1.3pre4";

    if (xmame.exists()) {
      try {
        String OS = System.getProperty("os.name").toLowerCase();
        // System.out.println(OS);
        if (OS.indexOf("windows") > -1) {
          cmd = "cmd.exe /c " + parsePath(xmamePath);
        }
        else {
          // Run command "xmame -V"
          cmd = xmamePath + " -V";
        }

        outlist = runCommand(cmd);

        for (int i = 0; i < outlist.length; i++) {
          // Get the line containing product version
	  if (outlist[i].startsWith("xmame") || 
              // Windows command line... 
              outlist[i].startsWith("M.A.M.E.")) {
            if (OS.indexOf("windows") > -1)
              version = outlist[i].substring(0, outlist[i].lastIndexOf(')') + 1);
            else
	      version = outlist[i];

            break;
          }
        }
      }
      catch (IOException e) {
      } // End try-catch
    }
          
    return version;
  }

  /**
   * This method returns true if a rom is correct, false otherwise.
   *
   * @param xmamePath The path to the xmame executable
   * @param romPath   The path to the rom to verify
   * @param romName   The name of the rom to verify
   *
   * @return          <code>true</code> if the rom is correct,
   *                  <code>false</code> otherwise
   *
   * @see             #verifyRom(String, String, String)
   *
   * @since           0.1.3pre2
   */
  public boolean romIsCorrect(String xmamePath, String romPath, String romName) {
    String response = verifyRom(xmamePath, romPath, romName);

    if (response.equals("romset " + romName + " correct") ||
        // Windows command line...
        response.equals("1 romsets found, 1 were OK."))
      return true;
    else
      return false;
  }

  /**
   * This method runs the command "xmame -verifyroms" to verify a romset
   * and gets the response.
   *
   * @param xmamePath The path to the xmame executable
   * @param romPath   The path to the rom to verify
   * @param romName   The name of the rom to verify
   *
   * @return          The <code>output</code> of the command
   *                  "xmame -verifyroms rom name".
   *
   * @see             #runCommand(String)
   */
  public String verifyRom(String xmamePath, String romPath, String romName) {
    String romStatus = new String();

    if (!xmamePath.equals("<NOT SET>")) {
      try {
        // Run command "xmame -verifyroms"
        String OS = System.getProperty("os.name").toLowerCase();
        // System.out.println(OS);
        if (OS.indexOf("windows") > -1) {
          cmd = getRealCommand(xmamePath) + " -verifyroms " + romName + " -rompath " +  romPath;
        }
        else {
          cmd = xmamePath  + " -rompath " + romPath + " -verifyroms " + romName;
        }

        
        outlist = runCommand(cmd);

        for (int i = 0; i < outlist.length; i++) {
          // Get the line we need
	  if (outlist[i].startsWith("romset") ||
              // Windows command line...
              outlist[i].startsWith("1 romsets"))
	    romStatus = outlist[i];
            break;
        }
      }
      catch (IOException e) {
        System.err.println(e);
      }
    }

    return romStatus;
  }

  /**
   * This method returns a string containing the number
   * of colors used by a rom running the command "xmame
   * -listcolors romname".
   *
   * @param xmamePath The path to the xmame executable
   * @param romName   The name of the rom
   *
   * @return          A <code>String</code> containing the number
   *                  of colors used by a given rom
   *
   * @see             #runCommand(String)
   */
  public String getRomColors(String xmamePath, String romName) {
    String colors = null;

    if (!xmamePath.equals("<NOT SET>")) {
      try {
        // Run command "xmame -listcolors"
        cmd = getRealCommand(xmamePath + " -listcolors " + romName);
        outlist = runCommand(cmd);

        for (int i = 0; i < outlist.length; i++) {
          // Get the line containing version info.
	  if (outlist[i].startsWith(romName))
	    colors = outlist[i].substring(10).trim();
        }
      }
      catch (IOException e) {
        System.err.println(e);
      }
    }
    return colors;
  }
  
  /**
   * This method adds available sound-dsp plugins into a
   * main window' s specific combo box.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @see             #getDigitalSoundPlugins(String)
   */
  public void setDigitalSoundPlugins(String xmamePath) {
    String[] dsp = getDigitalSoundPlugins(xmamePath);

    // Update main window' s dsp combobox with new items
    if (dsp.length > 0) {
      ActionListener[] al = romeo.options.dspBox.getActionListeners();

      for (int i = 0; i < al.length; i++) {
        romeo.options.dspBox.removeActionListener(al[i]);
      }

      // Get number of items in the list
      int count = romeo.options.dspBox.getItemCount();

      // If count > 0 removes all items
      if (count > 0)
        romeo.options.dspBox.removeAllItems();

      for (int i = 0; i < dsp.length; i++) {
	// Add item to the combo box
	romeo.options.dspBox.addItem(dsp[i]);
      }

      for (int i = 0; i < al.length; i++) {
        romeo.options.dspBox.addActionListener(al[i]);
      }
    }
  }

  /**
   * This method removes all unuseful stuff (ie copyright lines
   * (Open GL), etc.) from the list of the available sound-dsp
   * plugins.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of available sound-dsp plugins without
   *                  unuseful stuff
   *
   * @see             #listDigitalSoundPlugins(String)
   */
  public String[] getDigitalSoundPlugins(String xmamePath) {
    ArrayList plugins = new ArrayList();

    plugins.add("<NOT SET>");
    
    if (!xmamePath.equals("<NOT SET>")) {
      outlist = listDigitalSoundPlugins(xmamePath);

      for (int i = 0; i < outlist.length; i++) {
        // Do not add other things as copyright lines (Open GL) or....
        if (outlist[i].endsWith("plugin"))
	  plugins.add(outlist[i].substring(0, 20).trim());
      }
    }

    return (String[])plugins.toArray(new String[0]);
  }
  
  /**
   * This method lists available sound-dsp plugins running the 
   * command "xmame -ldp".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of available sound-dsp plugins
   *
   * @see             #runCommand(String)
   */
  public String[] listDigitalSoundPlugins(String xmamePath) {
    ArrayList ldpList = new ArrayList();

    try {
      // Run command 'xmame -ldp'. This command seems not to
      // terminate: its exit value is non-zero.
      cmd = getRealCommand(xmamePath + " -ldp -noloadconfig");
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++)
	ldpList.add(outlist[i].trim());
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])ldpList.toArray(new String[0]);
  }
  
  /**
   * This method adds available sound mixer plugins into a
   * main window' s specific combo box.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @see             #getSoundMixerPlugins(String)
   */
  public void setSoundMixerPlugins(String xmamePath) {
    String[] mixers = getSoundMixerPlugins(xmamePath);

    // Update main window' s mixer combobox with new items
    if (mixers.length > 0) {
      ActionListener[] al = romeo.options.dspBox.getActionListeners();

      for (int i = 0; i < al.length; i++) {
        romeo.options.mixerBox.removeActionListener(al[i]);
      }

      // Get number of items in the list
      int count = romeo.options.mixerBox.getItemCount();

      // If count > 0 removes all items
      if (count > 0)
        romeo.options.mixerBox.removeAllItems();

      for (int i = 0; i < mixers.length; i++) {
	// Add item to the combo box
	romeo.options.mixerBox.addItem(mixers[i]);
      }

      for (int i = 0; i < al.length; i++) {
        romeo.options.mixerBox.addActionListener(al[i]);
      }
    }
  }

  /**
   * This method removes all unuseful stuff (ie copyright lines
   * (Open GL), etc.) from the list of the available sound-mixer 
   * plugins.
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of available sound-mixer plugins without
   *                  unuseful stuff
   *
   * @see             #listSoundMixerPlugins(String)
   * @see             #runCommand(String)
   */
  public String[] getSoundMixerPlugins(String xmamePath) {
    ArrayList plugins = new ArrayList();

    plugins.add("<NOT SET>");
    
    if (!xmamePath.equals("<NOT SET>")) {
      // Run command 'xmame -ldp'
      outlist = listSoundMixerPlugins(xmamePath);

      for (int i = 0; i < outlist.length; i++) {
        // Do not add other things as copyright lines (Open GL) or....
        if (outlist[i].endsWith("plugin"))

	plugins.add(outlist[i].substring(0, 20).trim());
      }
    }

    return (String[])plugins.toArray(new String[0]);
  }

  /**
   * This method lists available sound-mixer plugins running the
   * command "xmame -lmp -noloadconfig".
   *
   * @param xmamePath The path to the xmame executable
   *
   * @return          The list of available sound-mixer plugins
   *
   * @see             #runCommand(String)
   */
  public String[] listSoundMixerPlugins(String xmamePath) {
    ArrayList ldpList = new ArrayList();

    try {
      // Run command 'xmame -ldp'. This command seems not to
      // terminate: its exit value is non-zero.
      cmd = getRealCommand(xmamePath + " -lmp -noloadconfig");
      outlist = runCommand(cmd);

      for (int i = 0; i < outlist.length; i++)
	ldpList.add(outlist[i].trim());
    }
    catch (IOException e) {
      System.err.println(e);
    }

    return (String[])ldpList.toArray(new String[0]);
  }
  
  /**
   * This method runs a given game.
   *
   * @param xmamePath The path to the xmame executable
   * @param options   The options selected by the user to run the given game
   *
   * @throws          IOException
   *
   * @see             #runCommand(String)
   */
  public void runMame(String xmamePath, String options) throws IOException {
    if (!xmamePath.equals("<NOT SET>")) {
      cmd = getRealCommand(xmamePath) + options;
      //System.out.println(cmd);

      try {
        // Run a game
        outlist = runCommand(cmd);

        // Display its output
        for (int i = 0; i < outlist.length; i++) {
	  romeo.games.appendMessage("Error", outlist[i]);
	}
      }
      catch (IOException e) {
        System.err.println(e);
      }
    }
  }

  /**
   * This method runs a given command, gets command' s output stream and
   * puts a buffered reader input stream on it, then reads the output lines
   * from command and put them into a list to pass to the caller.
   *
   * @param cmd The command to run
   *
   * @return    The list of strings for the caller
   *
   * @throws    IOException
   */
  static public String[] runCommand(String cmd) throws IOException {
    // Set up list to capture command output lines
    ArrayList list = new ArrayList();

    // Start command running
    Process proc = Runtime.getRuntime().exec(cmd);

    // Get command's output stream and
    // put a buffered reader input stream on it
    InputStream istr = proc.getInputStream();
    BufferedReader br = new BufferedReader(new InputStreamReader(istr));

    // Read output lines from command
    String str;
    while ((str = br.readLine()) != null) list.add(str);

    // Wait for command to terminate
    try {
      proc.waitFor();
    }
    catch (InterruptedException e) {
      System.err.println("Process was interrupted");
    }

    // Check its exit value
    if (proc.exitValue() != 0)
      if (!cmd.endsWith("-ldp -noloadconfig") && !cmd.endsWith("-lmp -noloadconfig"))
      System.err.println("Exit value was non-zero");

    // Close stream
    br.close();

    // Return list of strings to caller
    return (String[])list.toArray(new String[0]);
  }

  public String getRealCommand(String cmd) {
    String rcmd = new String();

    String OS = System.getProperty("os.name").toLowerCase();
    // System.out.println(OS);
    if (OS.indexOf("windows 9") > -1) {
      rcmd = "command.com /c " + parsePath(cmd);
    }
    else if ((OS.indexOf("nt") > -1) || 
             (OS.indexOf("windows 2000") > -1) || 
             (OS.indexOf("windows xp") > -1)) {
      rcmd = "cmd.exe /c " + parsePath(cmd);
      //System.out.println(rcmd);
    }
    else {
      // our last hope, we assume Unix
      rcmd = cmd;
    }

    return rcmd;
  }

  public String parsePath(String cmd) {
    // The end of the path containing chars "\"
    String partialPath;
    // Parsed string
    String parsedPath = new String();
    // The string returned at the end of this process
    String finalPath = new String();
    // First occurrence of the char "\"
    int idx_separator = cmd.indexOf(File.separator);
    //System.out.println("Char . alla posizione: " + idx_p);

    // Character "\" has been found
    if (idx_separator > 0) {
      parsedPath = cmd.substring(0, idx_separator);
      // Get last occurrence of char "\"
      int last_idx_separator = cmd.lastIndexOf(File.separator);

      // Recursively substitute all chars "\" with '\\'
      while (idx_separator <= last_idx_separator) {
	//System.out.println("Char \ alla posizione: " + idx_separator);
	partialPath = cmd.substring(idx_separator + 1);
	parsedPath = parsedPath + File.separator + File.separator + partialPath;
	int new_idx_separator = partialPath.indexOf(File.separator);
	if (new_idx_separator > 0) {
	  idx_separator = idx_separator + new_idx_separator + 1;
	  parsedPath = parsedPath.substring(0, idx_separator + 1);
	}
	// no more occurrences: break!
	else
	  break;
      }

      finalPath = parsedPath;
    }
    else {
      finalPath = cmd;
    }

    // return the correct path
    return finalPath;
  }
}  // That' s all, folks :)
